# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
PQDT, http://pqdtopen.proquest.com.

Config
------

[pqdt]

oai = http://example.endpoint.com/x/y/z

"""

import datetime

import luigi

from gluish.common import Executable
from gluish.format import Gzip
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.sources.amsl import AMSLFilterConfig
from siskin.task import DefaultTask


class PQDTTask(DefaultTask):
    """ PQDT open. """
    TAG = '34'

    def closest(self):
        return monthly(date=self.date)


class PQDTCombine(PQDTTask):
    """ Combine files."""

    date = ClosestDateParameter(default=datetime.date.today())
    prefix = luigi.Parameter(default="oai_dc")

    def requires(self):
        return Executable(name='metha-sync', message='https://github.com/miku/metha')

    def run(self):
        url = self.config.get('pqdt', 'oai')
        shellout("METHA_DIR={dir} metha-sync -format {prefix} {url}",
                 prefix=self.prefix,
                 url=url,
                 dir=self.config.get('core', 'metha-dir'))
        output = shellout("METHA_DIR={dir} metha-cat -root Records -format {prefix} {url} | pigz -c > {output}",
                          prefix=self.prefix,
                          url=url,
                          dir=self.config.get('core', 'metha-dir'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="xml.gz"))


class PQDTIntermediateSchema(PQDTTask):
    """
    OAI to intermediate schema.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return PQDTCombine(date=self.date)

    def run(self):
        mapdir = 'file:///%s' % self.assets("maps/")
        output = shellout("""flux.sh {flux} in={input} MAP_DIR={mapdir} | pigz -c > {output}""",
                          flux=self.assets("34/flux.flux"),
                          mapdir=mapdir,
                          input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class PQDTExport(PQDTTask):
    """
    OAI to SOLR schema.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='solr5vu3')

    def requires(self):
        return {
            'file': PQDTIntermediateSchema(date=self.date),
            'config': AMSLFilterConfig(date=self.date),
        }

    def run(self):
        output = shellout(""" span-tag -c {config} {input} > {output} """,
                          config=self.input().get('config').path,
                          input=self.input().get('file').path)
        output = shellout(""" span-export -o {format} {input} > {output} """, input=output, format=self.format)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))



import pymarc
from siskin.conversions import imslp_xml_to_marc


def test_imslp_xml_to_marc():
    example = """<?xml version="1.0"?>
    <document docID="imslpvalsskramstadhans">
      <localClass localClassName="col">imslp</localClass>
      <localClass localClassName="vifa">vifamusik</localClass>
      <identifier identifierEncodingSchema="originalID">valsskramstadhans</identifier>
      <creator>
        <mainForm>Skramstad, Hans</mainForm>
      </creator>
      <title>Vals for pianoforte</title>
      <subject>
        <mainForm>Romantic</mainForm>
      </subject>
      <music_arrangement_of>Piano</music_arrangement_of>
      <url urlEncodingSchema="originalDetailView">http://imslp.org/wiki/Vals_(Skramstad,_Hans)</url>
      <vifatype>Internetressource</vifatype>
      <fetchDate>2018-04-25T00:00:00.01Z</fetchDate>
      <vifaxml><![CDATA[<document docID="imslpvalsskramstadhans"><localClass
          localClassName="col">imslp</localClass><localClass
          localClassName="vifa">vifamusik</localClass><identifier
          identifierEncodingSchema="originalID">valsskramstadhans</identifier><creator><mainForm>Skramstad,
          Hans</mainForm></creator><title>Vals for
          pianoforte</title><subject><mainForm>Romantic</mainForm></subject><music_arrangement_of>Piano</music_arrangement_of><url
          urlEncodingSchema="originalDetailView">http://imslp.org/wiki/Vals_(Skramstad,_Hans)</url><vifatype>Internetressource</vifatype></document>]]></vifaxml>
    </document>
    """

    result = imslp_xml_to_marc(example)

    assert result is not None
    assert isinstance(result, pymarc.Record)

    assert result["001"].value() == "finc-15-dmFsc3NrcmFtc3RhZGhhbnM"
    assert result["100"]["a"] == "Skramstad, Hans"
    assert result["245"]["a"] == "Vals for pianoforte"
    assert result["856"]["u"] == "http://imslp.org/wiki/Vals_(Skramstad,_Hans)"



#!/usr/bin/env python3
# coding: utf-8

# SID: 20
# Ticket: #14793
# TechnicalCollectionID: sid-20-col-gallica, sid-20-col-gallicabuch
# Task: gallica


from __future__ import print_function

import re
import sys

import xmltodict

import marcx
from siskin.mappings import formats, roles
from siskin.utils import xmlstream, marc_build_field_008, check_isbn, check_issn

setlist = ["gallica:typedoc:partitions", "gallica:theme:0:00", "gallica:theme:0:02"]

inputfilename = "20_input.xml"
outputfilename = "20_output.mrc"

if len(sys.argv) == 3:
    inputfilename, outputfilename = sys.argv[1:]

outputfile = open(outputfilename, "wb")

for oldrecord in xmlstream(inputfilename, "Record"):
    
    oldrecord = xmltodict.parse(oldrecord, force_list=("setSpec", "dc:identifier", "dc:language", "dc:creator", "dc:title", "dc:publisher", "dc:rights", "dc:subject", "dc:relation", "dc:description"))
    marcrecord = marcx.Record(force_utf8=True)
    marcrecord.strict = False

    status = oldrecord["Record"]["header"]["@status"]
    setspec = oldrecord["Record"]["header"]["setSpec"][0]
    metadata = oldrecord["Record"]["metadata"]

    if setspec not in setlist or not metadata or status == "deleted":
        continue

    oldrecord = oldrecord["Record"]["metadata"]["ns0:dc"]

    # Formatmapping
    if setspec == "gallica:typedoc:partitions":
        format = "Score"
    else:
        format = "Book"

    # Leader
    leader = formats[format]["Leader"]
    marcrecord.leader = leader

    # Identifikator
    id = oldrecord["dc:identifier"][0]
    match = re.search("gallica.bnf.fr/ark:/\d+/(.*)", id)
    if match:
        f001 = match.group(1)
        f980a = match.group(1)
    else:
        sys.exit("Die ID konnte nicht ermittelt werden: " + f001)
    f001 = "finc-20-" + f001
    marcrecord.add("001", data=f001)

    # Zugangsart
    f007 = formats[format]["e007"]
    marcrecord.add("007", data=f007)

    # Periodizität
    year = oldrecord.get("dc:date", "")
    periodicity = formats[format]["008"]
    language = oldrecord.get("dc:language", [""])
    language = language[0]
    f008 = marc_build_field_008(year, periodicity, language)
    marcrecord.add("008", data=f008)

    # ISBN und ISSN
    numbers = oldrecord["dc:identifier"]
    for number in numbers:
        if "ISBN" in number:
            f020a = check_isbn(number)
            marcrecord.add("020", a=f020a)
        if "ISSN" in number:
            f022a = check_issn(number)
            marcrecord.add("022", a=f022a)
   
    # Sprache
    f041a = oldrecord.get("dc:language", [""])
    f041a = f041a[0]
    if f041a and len(f041a) != 3:
        sys.exit("Die Sprache ist nicht MARC-konform: " + f041a)
    marcrecord.add("041", a=f041a)

    # DDC-Notation
    if setspec == "gallica:theme:0:00":
        marcrecord.add("082", a="000")
    elif setspec == "gallica:theme:0:02":
        marcrecord.add("082", a="020")
    elif setspec == "gallica:typedoc:partitions":
        marcrecord.add("082", a="780")

    # 1. Schöpfer
    creators = oldrecord.get("dc:creator", "")
    if creators:
        match = re.search("(.*?)\s\(([0-9\.-]{9,9})\)\.\s(.*)", creators[0])
        if match:
            f100a, f100d, f100e = match.groups()
            f100e = f100e.lower()
            f1004 = roles.get(f100e, "")
            if not f1004:
                #print("Die Rollenbezeichnung '%s' fehlt in der Mapping-Tabelle." % f100e)
                pass
        else:
            f100a = creators[0]
            f100d = ""
            f1004 = ""
        marcrecord.add("100", a=f100a, d=f100d, _4=f1004)
   
    # Haupttitel, Titelzusatz und Verantwortliche
    titles = oldrecord["dc:title"]
    if len(titles) == 1:
        f245 = titles[0]
    else:
        if "#text" in titles[0] and "@xml:lang" in titles[0]:    #[OrderedDict([('@xml:lang', '')]), OrderedDict([('@xml:lang', ''), ('#text', 'Andante')])]
            f245 = titles[0]["#text"]
        elif "@xml:lang" in titles[0] and not "#text" in titles[0]:
            f245 = titles[1]["#text"]
        else:
            f245 = titles[0]
    match1 = re.search("(.*?)\s:\s(.*)\s\/\s(.*)", f245)
    match2 = re.search("(.*)\s:\s(.*)", f245)
    match3 = re.search("(.*)\s\/\s(.*)", f245)
    if match1:
        f245a = match1.group(1)
        f245b = match1.group(2)
        f245c = match1.group(3)
    elif match2:
        f245a = match2.group(1)
        f245b = match2.group(2)
        f245c = ""
    elif match3:
        f245a = match3.group(1)
        f245b = ""
        f245c = match3.group(2)
    else:
        f245a = f245
        f245b = ""
        f245c = ""
    subfields = ["a", f245a, "b", f245b, "c", f245c]
    marcrecord.add("245", subfields=subfields)
  
    # Alternativtitel
    if len(titles) == 2:
        if "@xml:lang" in titles[1]:
            f246a = titles[1]["#text"]
        else:
            f246a = titles[1]
        if f246a != f245:
            marcrecord.add("246", a=f246a)

    # Erscheinungsvermerk
    f260a = ""
    f260b = ""
    publishers = oldrecord.get("dc:publisher", "")
    for publisher in publishers:
        match = re.search("(.*)\s\((.*)\)", publisher)
        if match:
            f260a, f260b = match.groups()
        else:
            f260a = publisher
            f260b = ""
        f260c = oldrecord.get("dc:date", "") 
        subfields = ["a", f260a, "b", f260b, "c", f260c]
        marcrecord.add("260", subfields=subfields)

    # Umfangsangabe
    f300a = oldrecord.get("dc:format", "")
    marcrecord.add("300", a=f300a)

    # RDA-Inhaltstyp
    f336b = formats[format]["336b"]
    marcrecord.add("336", b=f336b)

    # RDA-Datenträgertyp
    f338b = formats[format]["338b"]
    marcrecord.add("338", b=f338b)

    # Rechtehinweis
    f500a = oldrecord.get("dc:rights", [""])
    if len(f500a) == 2:
        f500a = f500a[1]["#text"].title()
    else:
        f500a = f500a[0].title()
    marcrecord.add("500", a=f500a)

    # Inhaltsbeschreibung
    descriptions = oldrecord.get("dc:description", [""])
    for f520a in descriptions:
        if f520a:   # sometimes None
            if len(f520a) > 8000:
                f520a = f520a[:8000]
            marcrecord.add("520", a=f520a)

    # Schlagwort
    subjects = oldrecord.get("dc:subject", "")
    for subject in subjects:
        f650a = subject["#text"]
        marcrecord.add("650", a=f650a)

    # GND-Inhalts- und Datenträgertyp
    f655a = formats[format]["655a"]
    f6552 = formats[format]["6552"]
    marcrecord.add("338", a=f655a, _2=f6552)

    # weitere geistige Schöpfer
    creators = oldrecord.get("dc:creator", "")
    for creator in creators[1:]:
        match = re.search("(.*?)\s\(([0-9\.-]{9,9})\)\.\s(.*)", creator)
        if match:
            f700a, f700d, f700e = match.groups()
            f700e = f700e.lower()
            f7004 = roles.get(f700e, "")
            if not f7004:
                #print("Die Rollenbezeichnung '%s' fehlt in der Mapping-Tabelle." % f700e)
                pass
        else:
            f700a = creator
            f700d = ""
            f7004 = ""
        marcrecord.add("700", a=f700a, d=f700d, _4=f7004)

    # Link zu Datensatz und Ressource
    f856u = oldrecord["dc:relation"][0]
    match = re.search("(http.*catalogue.*)", f856u)  # Notice du catalogue : http://catalogue.bnf.fr/ark:/12148/cb42874085r
    if match:
        f856u = match.group(1)
        marcrecord.add("856", q="text/html", _3="Link zum Datensatz", u=f856u)
    
    f856u = oldrecord["dc:identifier"][0]
    marcrecord.add("856", q="image/jpeg", _3="Link zum Digitalisat", u=f856u)

    # SWB-Inhaltstyp
    f935c = formats[format]["935c"]
    marcrecord.add("935", c=f935c)

    # Ansigelung
    if setspec == "gallica:theme:0:00":
        f980c = "sid-20-col-gallicabuch"
    elif setspec == "gallica:theme:0:02":
        f980c = "sid-20-col-gallicabuch"
    else:
        f980c = "sid-20-col-gallica"
    collections = ["a", f980a, "b", "20", "c", f980c]
    marcrecord.add("980", subfields=collections)

    outputfile.write(marcrecord.as_marc())

outputfile.close()


#!/usr/bin/env python3
# coding: utf-8

# Quelle: VDEH
# SID: 130
# Tickets: #9868, #15025
# technicalCollectionID: sid-130-col-vdeh
# Task: vdeh.py

from __future__ import print_function

import collections
import io
import re
import sys
from builtins import *

import requests
import xmltodict

import marcx
from siskin.mab import MabXMLFile
from siskin.utils import marc_build_imprint

formatmap = {
    "p||||||z|||||||": u"Zeitschrift",
    "z||||||z|||||||": u"Zeitung",
    "r||||||z|||||||": u"Reihe",
    "r||||||||||||||": u"Reihe",
    "a|a|||||||||||": u"Monografie",
    "a|||||||||||||": u"Monografie",
    "a|a|||||||": u"Monografie",
    "|||||ca||||||": u"Videokassette",
    "a|||||||g|||||": u"Datenträger",
    "||||||||g|": u"Datenträger"
}


def check_swb_isbn(isbn):
    """
    The length of the response can indicate, whether records exist.
    """
    req = requests.get("%s?q=source_id:0+institution:DE-105+isbn:%s&wt=csv&fl=id" % (servername, isbn))
    return len(req.text)


def check_swb_issn(issn, title):
    """
    The length of the response can indicate, whether records exist.
    """
    req = requests.get('%s?q=source_id:0+institution:DE-105+issn:%s+title_short:"%s"&wt=csv&fl=id' %
                       (servername, issn, title))
    return len(req.text)


inputfilename = "130_input.xml"
outputfilename = "130_output.mrc"
servername = "https://index.ub.uni-leipzig.de/solr/biblio/select"

if len(sys.argv) == 4:
    inputfilename, outputfilename, servername = sys.argv[1:]

reader = MabXMLFile(inputfilename, replace=((u"¬", ""), ))
outputfile = open(outputfilename, "wb")

hierarchymap = collections.defaultdict(list)

for record in reader:

    f010 = ""
    f089 = ""

    f010 = record.field("010")
    f089 = record.field("089")

    if f010 and f089:
        f010 = f010.lstrip("0")
        hierarchymap[f010].append(f089)

for record in reader:

    marcrecord = marcx.Record(force_utf8=True)
    marcrecord.strict = False

    subjects = []
    persons = {}
    corporates = []

    # Formatmapping
    format1 = record.field("052", alt="")
    format2 = record.field("050", alt="")

    if format1 != "" or format2 != "":
        format = format1 or format2
    else:
        format = "a|a|||||||||||"

    if formatmap.get(format) == "Zeitschrift":
        leader = "     cas a2202221   4500"
        f007 = "tu"
        f008 = "                     p"
        f935b = "druck"
        f935c = "az"
    elif formatmap.get(format) == "Zeitung":
        leader = "     cas a2202221   4500"
        f007 = "tu"
        f008 = "                     n"
        f935b = "druck"
        f935c = "az"
    elif formatmap.get(format) == "Reihe":
        leader = "     cas a2202221   4500"
        f007 = "tu"
        f008 = "                     m"
        f935b = "druck"
        f935c = "az"
    elif formatmap.get(format) == "Monografie":
        leader = "     nam  22        4500"
        f007 = "tu"
        f008 = ""
        f935b = "druck"
        f935c = ""
    elif formatmap.get(format) == "Videokassette":
        leader = "     cgm  22        4500"
        f007 = "v"
        f008 = ""
        f935b = "vika"
        f935c = "vide"
    elif formatmap.get(format) == "Datenträger":
        leader = "     cgm  22        4500"
        f007 = "v"
        f008 = ""
        f935b = "soerd"
        f935c = ""
    else:
        leader = "     nam  22        4500"
        f007 = "tu"
        f008 = ""
        f935b = "druck"
        f935c = ""

    assert (len(leader) == 24)
    marcrecord.leader = leader

    # Identfikator
    f001 = record.field("001", alt="")
    if f001 == "":
        continue
    marcrecord.add("001", data="finc-130-" + f001)

    # Format
    marcrecord.add("007", data=f007)
    marcrecord.add("008", data=f008)

    # ISBN
    f020a = record.field("540", alt="")
    if f020a:
        match = re.search("([0-9xX-]{10,17})", f020a)
        if match:
            f020a = match.group(1)
            x = check_swb_isbn(f020a)
            if x > 3:
                continue
            marcrecord.add("020", a=f020a)

    # ISSN
    f022a = record.field("542", alt="")
    if f022a:
        match = re.search("([0-9xX-]{9,9})", f022a)
        if match:
            f022a = match.group(1)
            x = check_swb_issn(f022a, f245a)
            if x > 3:
                continue
            marcrecord.add("022", a=f022a)

    # Sprache
    languages = record.field("037", alt="")
    f041a = [languages[i:i + 3]
             for i in range(0, len(languages), 3)]  # to handle uncommon language codes like "gerengfre"
    marcrecord.add("041", a=f041a)

    # 1. Urheber
    f100a = record.field("100", alt="")
    f100e = ""
    if f100a:
        match = re.search(u".\s(\[.*\])", f100a)
        if match:
            f100e = match.group(1)
            f100a = re.sub(u".\s\[.*\]", "", f100a)
        marcrecord.add("100", a=f100a, e=f100e)

    # 1. Körperschaft
    f110a = record.field("200", alt="")
    marcrecord.add("110", a=f110a)

    # Titel und Verantwortlichenangabe
    f245a = record.field("331", alt="")
    f245b = record.field("335", alt="")
    f245c = record.field("359", alt="")
    if len(f245a) < 3 or "Arkady" in f245a:
        # einzelne Zeitschriftenhefte, Titel nur aus Komma und Leerzeichen und die fehlerhaften Arkady-Records werden übersprungen
        continue
    titleparts = ["a", f245a, "b", f245b, "c", f245c]
    marcrecord.add("245", subfields=titleparts)

    # Ausgabe
    f250a = record.field("403", alt="")
    marcrecord.add("250", a=f250a)

    # Erscheinungsvermerk
    f260a = record.field("410", alt="")
    if not f260a:
        f260a = record.field("419", "a", alt="")

    f260b = record.field("412", alt="")
    if not f260b:
        f260b = record.field("419", "b", alt="")

    f260c = record.field("425", alt="")
    if not f260c:
        f260c = record.field("419", "c", alt="")

    subfields = marc_build_imprint(f260a, f260b, f260c)
    marcrecord.add("260", subfields=subfields)

    # Umfang
    f300a = record.field("433", alt="")
    f300b = record.field("434", alt="")
    if f300a and f300b:
        f300b = " : " + f300b
    f300c = record.field("435")
    if (f300a or f300b) and f300c:
        f300c = " ; " + f300c
    physicaldescription = ["a", f300a, "b", f300b, "c", f300c]
    marcrecord.add("300", subfields=physicaldescription)

    # Schlagwörter
    subjects = record.fields("710")
    for f650a in subjects:
        marcrecord.add("650", a=f650a)

    # weitere Personen
    for i in range(101, 197):
        f700a = record.field(str(i), alt="")
        match = re.search("^\d+", f700a)  # die Felder, die nur Personen-IDs enthalten, werden übersprungen
        if not match:
            match = re.search(".\s(\[.*\])", f700a)
            if match:
                f700e = match.group(1)
                f700a = re.sub(".\s\[.*\]", "", f700a)
            else:
                f700e = ""
            persons[f700a] = f700e

    for f700a, f700e in persons.items():
        marcrecord.add("700", a=f700a, e=f700e)

    # weitere Körperschaften
    for i in range(201, 299):
        f710a = record.field(i, alt="")
        match = re.search("^\d+", f710a)  # die Felder, die nur Körperschafts-IDs enthalten, werden übersprungen
        if not match:
            corporates.append(f710a)

    for f710a in corporates:
        marcrecord.add("710", a=f710a)

    # Link zum Datensatz
    f856u = record.field("655", "u", alt="")
    if f856u:
        marcrecord.add("856", q="text/html", u=f856u)

    #Bestandsnachweis (866a)
    f866a = f001.lstrip("0")
    f866a = hierarchymap.get(f866a, "")
    f866a = "; ".join(f866a)
    if not isinstance(f866a, str):
        f866a = f866a.decode('utf-8')
    marcrecord.add("866", a=f866a)

    # SWB-Format
    marcrecord.add("935", b=f935b, c=f935c)

    # Kollektion
    marcrecord.add("980", a=f001, b="130", c="sid-130-col-vdeh")

    outputfile.write(marcrecord.as_marc())

outputfile.close()



#!/usr/bin/env python3
# coding: utf-8
# pylint: disable=C0103

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Robert Schenk, <robert.schenk@uni-leipzig.de>
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar. If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>


# Contains a collection of utilities for parsing and analysing xml files
# Current functions: list_fields, count_records, print_json, list_all_values, list_unique_values
# To do: sort tags, count_doublet

from __future__ import print_function

import argparse
import collections
import io
import json
import os
import re
import subprocess
import xml.sax


class FieldExampleHandler(xml.sax.ContentHandler):
    """
    This handler collects example values for XML tags.
    """

    def __init__(self):
        self.examples = {}
        self.current = (None, None)

    def startElement(self, name, attr):
        self.current = (name, attr)

    def characters(self, value):
        name = self.current[0]
        value = value.strip()
        if value:
            if name not in self.examples:
                self.examples[name] = value


class StreamingJsonHandler(xml.sax.ContentHandler):
    """
    Read XML and emit JSON in streaming fashion.
    """

    def __init__(self):
        self.data = collections.defaultdict(list)
        self.current = (None, None)

    def startElement(self, name, attr):
        self.current = (name, attr)

    def endElement(self, name):
        print(json.dumps(self.data))
        self.data.clear()

    def characters(self, value):
        name = self.current[0]
        value = value.strip()
        if value:
            self.data[name].append(value)


class CountingHandler(xml.sax.ContentHandler):
    """
    This handler counts the number of records (by tag name).
    """

    def __init__(self, tag):
        self.count = 0
        self.tag = tag
        self.start_element_counter = 0

    def startElement(self, name, attr):
            if not self.tag:
                self.start_element_counter += 1
                if self.start_element_counter == 2:
                    self.tag = name

    def endElement(self, name):
        if name == self.tag:
            self.count += 1


class ListAllValuesHandler(xml.sax.ContentHandler):
    """
    This handler list all values of a given field.
    """

    def __init__(self, tag):
        self.tag = tag
        self.current = None

    def startElement(self, name, attr):
        self.current = name

    def characters(self, value):
        name = self.current
        if name == self.tag:
            value = value.strip()
            if value:
                print(value)


class ListUniqueValuesHandler(xml.sax.ContentHandler):
    """
    This handler list all unique values of a given field.
    """

    def __init__(self, tag):
        self.tag = tag
        self.current_tag = None
        self.current_values = [] # Collect chars within a tag here.
        self.seen = set() # All values, that have been printed already.

    def startElement(self, name, attr):
        self.current_tag = name

    def characters(self, value):
        if self.current_tag == self.tag:
            self.current_values.append(value)

    def endElement(self, name):
        if name == self.tag:
            value = ''.join(self.current_values)
            if value not in self.seen: 
                print(value)
                self.seen.add(value)
            self.current_values = []

parser = argparse.ArgumentParser()
parser.add_argument("-v",
                    action="version",
                    help="show version",
                    version="0.2.3")
parser.add_argument("-f",
                    dest="inputfile",
                    help="file to parse",
                    metavar="filename")
parser.add_argument("-t",
                    dest="task",
                    help="task to do (list_fields, count_records, print_json, list_all_values, list_unique_values)",
                    metavar="task")
parser.add_argument("-tag",
                    dest="tag",
                    help="record tag name used for counting",
                    metavar="tag")

args = parser.parse_args()
parser = xml.sax.make_parser()

if args.task == "list_fields":
    handler = FieldExampleHandler()
    parser.setContentHandler(handler)
    with open(args.inputfile, "r") as handle:
        parser.parse(handle)
    for key, value in handler.examples.items():
        print(key + " = " + value)

elif args.task == "print_json":
    handler = StreamingJsonHandler()
    parser.setContentHandler(handler)
    with open(args.inputfile, "r") as handle:
        parser.parse(handle)

elif args.task == "count_records":
    handler = CountingHandler(args.tag)
    parser.setContentHandler(handler)
    with open(args.inputfile, "r") as handle:
        parser.parse(handle)
    print(handler.count)

elif args.task == "list_all_values":
    handler = ListAllValuesHandler(args.tag)
    parser.setContentHandler(handler)
    with open(args.inputfile, "r") as handle:
        parser.parse(handle)

elif args.task == "list_unique_values":
    handler = ListUniqueValuesHandler(args.tag)
    parser.setContentHandler(handler)
    with open(args.inputfile, "r") as handle:
        parser.parse(handle)



#!/usr/bin/env python3
# coding: utf-8

import io
import json
import re
import sys

import marcx


def get_field(tag):
    try:
        value = jsonrecord[tag]
    except:
        value = ""
    return value


inputfilename = "10_input.json"
outputfilename = "10_output.mrc"

if len(sys.argv) == 3:
    inputfilename, outputfilename = sys.argv[1:]

outputfile = io.open(outputfilename, "wb")
inputfile = open(inputfilename, "r")

for line in inputfile:
    jsonobject = json.loads(line)
    jsonrecords = jsonobject["content"]["results"]

    for jsonrecord in jsonrecords:

        marcrecord = marcx.Record(force_utf8=True)
        marcrecord.strict = False

        format = get_field("original_format")
        if "notated music" in format:
            leader = "     com  22        4500"
            f935c = "muno"
        elif "web page" in format:
            leader = "     nmi  22        4500"
            f935c = "website"
        #elif "audio recording" in format:
        #    leader = "     nmi  22        4500"
        #    f935c = "muto"
        #elif "collection" in format:
        #    leader = "     nom  22        4500"
        #    f935c = ""
        elif "photo" in format:
            leader = "     ckm  22        4500"
            f935c = "foto"
        else:
            leader = "     cam  22        4500"
            f935c = "text"

        # Leader
        marcrecord.leader = leader

        # Format
        marcrecord.add("007", data="cr")

        #Identifikator
        f001 = jsonrecord["id"]
        f001 = f001.rstrip("/").split("/")[-1]
        marcrecord.add("001", data="finc-10-" + f001)

        # 1. Urheber
        f100a = get_field("contributor")
        if f100a != "":
            f100a = f100a[0]
            f100a = f100a.title()
            marcrecord.add("100", a=f100a)

        # Haupttitel
        f245a = get_field("title")
        marcrecord.add("245", a=f245a)

        # Alternativtitel
        atitles = get_field("other_title")
        for atitle in atitles:
            marcrecord.add("246", a=atitle)

        # Erscheinungsjahr
        f260c = get_field("date")
        marcrecord.add("260", c=f260c)

        # Schlagwörter
        subjects = get_field("subject")
        for subject in subjects:
            subject = subject.title()
            marcrecord.add("650", a=subject)

        # weitere Urheber
        persons = get_field("contributor")
        if persons != "":
            for person in persons[1:]:
                person = person.title()
                marcrecord.add("700", a=person)

        #n Link zum Datensatz
        f856u = get_field("url")
        marcrecord.add("856", q="text/html", _3="Link zur Ressource", u=f856u)

        # Format
        marcrecord.add("935", b="cofz", c=f935c)

        # Kollektion
        marcrecord.add("980", a=f001, b="10", c="Music Treasures Consortium")

        outputfile.write(marcrecord.as_marc())

inputfile.close()
outputfile.close()



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Robert Schenk, <robert.schenk@uni-leipzig.de>
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
IMSLP, refs #1240, #13055.

Config
------

[imslp]

listings-url = http://example.com/export/
legacy-mapping = /path/to/json.file
"""

import datetime
import json
import os
import shutil
import tempfile

import luigi

from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.conversions import imslp_tarball_to_marc
from siskin.decorator import deprecated
from siskin.task import DefaultTask
from siskin.utils import scrape_html_listing


class IMSLPTask(DefaultTask):
    """ Base task for IMSLP. """
    TAG = "15"

    def closest(self):
        return datetime.date(2018, 4, 25)

    def latest_link(self):
        """
        Find the lastest link on the listing.
        """
        listings_url = self.config.get("imslp", "listings-url")
        links = sorted([link for link in scrape_html_listing(listings_url) if "imslpOut_" in link])

        self.logger.debug("found %s links on IMSLP download site", len(links))

        if len(links) == 0:
            raise ValueError("could not find any suitable links: %s", listings_url)

        self.logger.debug("assuming latest link is %s", links[-1])
        return links[-1]


class IMSLPDownloadDeprecated(IMSLPTask):
    """
    Download raw data. Should be a single URL pointing to a tar.gz.

    Deprecated, will be removed soon.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    @deprecated
    def run(self):
        output = shellout("""wget -O {output} {url}""", url=self.config.get('imslp', 'url'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='tar.gz'))


class IMSLPDownload(IMSLPTask):
    """
    Download raw data. Actually keeps the original filename. Should be a tar.gz.
    """

    date = luigi.DateParameter(default=datetime.date.today())

    def run(self):
        """
        Assume, links are sortable, the last item should contain the latest link.
        XXX: Do not download, if not newer than last downloaded file.
        """
        most_recent = self.latest_link()
        self.logger.debug("most recent package seem to be: %s", most_recent)
        output = shellout(""" wget -O {output} "{url}" """, url=most_recent)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        filename = os.path.basename(self.latest_link())
        dst = os.path.join(self.taskdir(), filename)
        return luigi.LocalTarget(path=dst)


class IMSLPLegacyMapping(IMSLPTask, luigi.ExternalTask):
    """
    Path to JSON file mapping record id to viaf id and worktitle ("Werktitel").

    This is a fixed file.

        {
            "fantasiano10mudarraalonso": {
                "title": "Fantasie No.10",
                "viaf": "(VIAF)100203803"
            },
            ...
        }
    """

    def output(self):
        return luigi.LocalTarget(path=self.config.get('imslp', 'legacy-mapping'))


class IMSLPConvertNext(IMSLPTask):
    """
    Take a current version of the data plus legacy mapping and convert.

    WIP, refs #12288, refs #13055. May merge with 15_marcbinary.py.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'legacy-mapping': IMSLPLegacyMapping(),
            'data': IMSLPDownload(date=self.date),
        }

    def run(self):
        """
        Load mapping, convert, write.
        """
        with self.input().get("legacy-mapping").open() as handle:
            mapping = json.load(handle)

        output = imslp_tarball_to_marc(self.input().get("data").path, legacy_mapping=mapping)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        filename = os.path.basename(self.latest_link())
        dst = os.path.join(self.taskdir(), filename.replace("tar.gz", "fincmarc.mrc"))
        return luigi.LocalTarget(path=dst)


class IMSLPConvert(IMSLPTask):
    """
    Extract and transform.

    TODO, refs #13055 -- see IMSLPDownloadNext and IMSLPConvertNext and IMSLPLegacyMapping.
    """

    date = ClosestDateParameter(default=datetime.date.today())
    debug = luigi.BoolParameter(description='do not delete temporary folder', significant=False)

    def requires(self):
        return IMSLPDownload(date=self.date)

    def run(self):
        tempdir = tempfile.mkdtemp(prefix='siskin-')
        shellout("tar -xzf {archive} -C {tempdir}", archive=self.input().path, tempdir=tempdir)
        output = shellout("python {script} {tempdir} {output} {fieldmap}",
                          script=self.assets('15/15_marcbinary.py'),
                          tempdir=tempdir,
                          fieldmap=self.assets('15/15_fieldmap.json'))
        if not self.debug:
            shutil.rmtree(tempdir)
        else:
            self.logger.debug("not deleting temporary folder at %s", tempdir)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='fincmarc.mrc'))



#!/usr/bin/env python
# coding: utf-8

from builtins import *

import xlrd

import marcx

# Default input and output.
inputfilename = "100_Medienwissenschaft_Berichte_Papiere.xlsx"
outputfilename = "100_output.mrc"

if len(sys.argv) == 3:
    inputfilename, outputfilename = sys.argv[1:]

outputfile = open(outputfilename, "wb")

workbook = xlrd.open_workbook(inputfilename)
sheet = workbook.sheet_by_name("Tabelle1")

for i, row in enumerate(range(sheet.nrows), start=0):

    csv_record = sheet.row_values(row)
    marc_record = marcx.Record(force_utf8=True)

    if csv_record[0] == "authors":
        continue

    # Leader
    marc_record.leader = "     naa  22        4500"

    # ID
    f001 = str(i)
    marc_record.add("001", data="finc-100-" + f001)

    # 007
    marc_record.add("007", data="cr")

    #008
    f260c = csv_record[8]
    f260c = str(f260c).rstrip(".0")
    if len(f260c) == 4:
        marc_record.add("008", data="130227u%suuuuuxx uuup     c" % f260c)
    else:
        marc_record.add("008", data="130227uu20uuuuuuxx uuup     c")

    # ISSN
    f022a = csv_record[5]
    marc_record.add("022", a=f022a)

    # 1. Schöpfer
    f100a = csv_record[0]
    f100a = f100a.split("; ")
    f100a = f100a[0]
    marc_record.strict = False
    marc_record.add("100", a=f100a)
    marc_record.strict = True

    # Haupttitel
    f245a = csv_record[1]
    f245a = str(f245a)
    marc_record.add("245", a=f245a)

    # Verlag, Erscheinungsjahr
    f260b = csv_record[3]
    f260b = str(f260b)
    f260b = f260b.lstrip("Hamburg: ")
    f260c = csv_record[8]
    f260c = str(f260c).rstrip(".0")
    publisher = ["a", "Hamburg : ", "b", f260b + ", ", "c", f260c]
    marc_record.add("260", subfields=publisher)

    # Seitenzahl
    f300a = csv_record[11]
    f300a = str(f300a).rstrip(".0")
    marc_record.add("300", a=f300a)

    # weitere Schöpfer
    f700a = csv_record[0]
    f700a = f700a.split("; ")
    if len(f700a) > 1:
        for person in f700a[1:]:
            marc_record.add("700", a=person)

    # Quelle
    f773t = csv_record[2]
    f773t = str(f773t)
    issue = csv_record[6]
    issue = str(issue).rstrip(".0")
    year = csv_record[8]
    year = str(year).rstrip(".0")
    pages = csv_record[12]
    pagess = str(pages).rstrip(".0")
    f773g = "(%s)%s, S. %s" % (year, issue, pages)
    marc_record.add("773", g=f773g, t=f773t)

    # Link zur Ressource
    f856u = csv_record[13]
    marc_record.add("856", q="text/html", _3="Link zur Ressource", u=f856u)

    # Medienform
    marc_record.add("935", b="cofz")

    # Kollektion
    values = ["a", f001, "b", "100", "c", "sid-100-col-medienwiss"]
    marc_record.add("980", subfields=values)

    outputfile.write(marc_record.as_marc())

outputfile.close()



# coding: utf-8
# pylint: disable=W0232,E1101,R0904,E1103,C0301

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Robert Schenk, <robert.schenk@uni-leipzig.de>
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#
"""
B3Kat, #8697.

The Bavarian Library Network (https://www.bib-bvb.de/) hosts B3Kat (around 228 institutions in 2018).

> Der B3Kat ist die gemeinsame Katalogisierungsplattform der
Bibliotheksverbünde BVB und KOBV. Die Hochschulbibliotheken der drei
Bundesländer Bayern, Berlin und Brandenburg, die Bayerische Staatsbibliothek
sowie zahlreiche Spezialbibliotheken rücken mit dem Aufbau des B3Kat näher
zusammen. Der B3Kat bildet das Kernstück der Strategischen Allianz des BVB
zusammen mit dem KOBV.

```
>>> import pandas as pd
>>> pd.read_html("http://www.gateway-bayern.de/bvb_biblist_sort_ger.html")[0].head()
Typ    ISIL Sigel Kürzel                                    Bibliotheksname         Ort Recherche
0    1   DE-12    12  BSB01               Bayerische  Staatsbibliothek München     München   Katalog
1    2  DE-384   384  UBA01                    Universitätsbibliothek Augsburg    Augsburg   Katalog
2    2  DE-473   473  UBG01                    Universitätsbibliothek  Bamberg     Bamberg   Katalog
3    2  DE-703   703  UBT01                   Universitätsbibliothek  Bayreuth    Bayreuth   Katalog
4    2  DE-945   945  UEI03  Universitätsbibliothek Eichstätt -  Wirtschaft...  Ingolstadt   Katalog

>>> df.groupby('Ort').size().sort_values(ascending=False).head(15)
Ort
München       69
Regensburg    19
Berlin        19
Augsburg      15
Bamberg        7
Würzburg       6
Ingolstadt     6
Nürnberg       6
Hof            4
Freising       4
Passau         3
Erlangen       3
Coburg         3
Fürth          3
Ansbach        3
dtype: int64
```

"""

import datetime
import os
import tempfile

import luigi

from gluish.format import TSV, Gzip
from gluish.intervals import semiyearly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.task import DefaultTask


class B3KatTask(DefaultTask):
    """
    Base task for B3Kat.
    """
    TAG = 'b3kat'

    def closest(self):
        """ Find exact date on this site: https://www.bib-bvb.de/web/b3kat/open-data """
        return semiyearly(date=self.date)


class B3KatLinks(B3KatTask):
    """
    A list of latest links. Currently grepped from https://www.bib-bvb.de/web/b3kat/open-data.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def run(self):
        output = shellout("""
            curl --fail -s "https://www.bib-bvb.de/web/b3kat/open-data" |
            grep -Eo "/OpenData/b3kat_export_[0-9]{{4,4}}_[0-9]{{1,2}}_teil[0-9]{{1,2}}.xml.gz" |
            awk '{{print "https://www.bib-bvb.de"$0 }}' > {output} """)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class B3KatDownload(B3KatTask):
    """
    Download snapshot. Output is a single (large) MARC binary file. Typically
    the downloads are provided in May and November.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return B3KatLinks(date=self.date)

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with self.input().open() as handle:
            for i, row in enumerate(handle.iter_tsv(cols=('url', )), start=1):
                downloaded = shellout("""curl -sL --fail "{url}" > {output} """, url=row.url)
                output = shellout("""yaz-marcdump -i marcxml -o marc "{input}" >> {stopover}""",
                                  input=downloaded,
                                  stopover=stopover)
                try:
                    os.remove(downloaded)
                    os.remove(output)
                except OSError as err:
                    self.logger.error(err)

        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))


class B3KatMARCXML(B3KatTask):
    """
    Convert binary MARC to MARCXML.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return B3KatDownload(date=self.date)

    def run(self):
        output = shellout("yaz-marcdump -i marc -o marcxml {input} | pigz -c > {output}", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml.gz'), format=Gzip)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Robert Schenk, <robert.schenk@uni-leipzig.de>
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
Persee, refs #3133, #11349.

# Config

[persee]

# one issn per line
fid-issn-url = http://example.com/fid-issn-list
"""

import datetime
import os

import luigi
import requests
import ujson as json

import marcx
import pymarc
from gluish.format import TSV
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.sources.amsl import AMSLService
from siskin.task import DefaultTask


class PerseeTask(DefaultTask):
    """ Base task. """
    TAG = '39'

    def closest(self):
        return monthly(self.date)


class PerseeCombined(PerseeTask):
    """
    Harvest and combine.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def run(self):
        """
        Explicit from, because earliestDate might be off, refs #11349
        (#note-65). In doubt, delete the cache via `rm -rf $(metha-sync -dir
        http://oai.persee.fr/oai)`.
        """
        output = shellout("""
            METHA_DIR={metha_dir} metha-sync -format marc -from 2000-01-01 http://oai.persee.fr/oai &&
            METHA_DIR={metha_dir} metha-cat -format marc http://oai.persee.fr/oai | pigz -c > {output}
        """,
                          metha_dir=self.config.get('core', 'metha-dir'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml.gz'))


class PerseeMARC(PerseeTask):
    """
    Custom convert to MARC.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return PerseeCombined(date=self.date)

    def run(self):
        issnfile = shellout(""" curl -sL --fail "{fid_issn_url}" > {output} """,
                            fid_issn_url=self.config.get('persee', 'fid-issn-url'))
        output = shellout("""python {script} <(unpigz -c {input}) {output} {issnfile}""",
                          script=self.assets('39/39_marcbinary.py'),
                          input=self.input().path,
                          issnfile=issnfile)
        luigi.LocalTarget(output).move(self.output().path)
        os.remove(issnfile)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='fincmarc.mrc'))

    def on_success(self):
        """
        Signal for manual update. This is decoupled from the task output,
        because it might be deleted, when imported.
        """
        self.output().copy(self.path(ext='fincmarc.mrc.import'))



#!/usr/bin/env python3
# coding: utf-8

# SID: 150
# Ticket: #11832, #15175
# TechnicalCollectionID: sid-150-col-monami
# Task: HSMW

import io
import re
import sys

import xmltodict

import marcx


def ddcmatch(value):
    """
    Returns true, if given value matches profile.
    """
    if value in ("091", "092", "093", "095"):
        return True

    for prefix in ("002", "020", "021", "022", "023", "025", "026", "027", "028", "070", "090", "094", "760"):
        if value.startswith(prefix):
            return True
    return False


inputfilename = "150_input.xml"
outputfilename = "150_output.mrc"

if len(sys.argv) == 3:
    inputfilename, outputfilename = sys.argv[1:3]

inputfile = open(inputfilename, "rb")
outputfile = open(outputfilename, "wb")

xmlfile = inputfile.read()
xmlrecords = xmltodict.parse(xmlfile, force_list="dc:subject")

for xmlrecord in xmlrecords["Records"]["Record"]:

    try:
        f245 = xmlrecord["metadata"]["oai_dc:dc"]["dc:title"]
    except:
        continue

    # Closed Access werden übersprungen
    access = xmlrecord["metadata"]["oai_dc:dc"]["dc:rights"][1]
    if access == "info:eu-repo/semantics/closedAccess":
        continue

    marcrecord = marcx.Record(force_utf8=True)
    marcrecord.strict = False

    # Leader
    marcrecord.leader = "     cam  22        4500"

    # Identifier
    f001 = xmlrecord["header"]["identifier"]
    regexp = re.match("oai:opus.bsz-bw.de-hsmw:(\d+)", f001)
    regexp2 = re.match("oai:monami.hs-mittweida.de:(\d+)", f001)
    if regexp:
        f001 = regexp.group(1)
    elif regexp2:
        f001 = regexp2.group(1)
    else:
        sys.exit("Keine ID vorhanden!" + f001)
    marcrecord.add("001", data="finc-150-" + f001)

    # 007
    marcrecord.add("007", data="cr")

    # Sprache
    language = xmlrecord["metadata"]["oai_dc:dc"]["dc:language"]
    if language == "deu":
        language = "ger"
    marcrecord.add("008", data="130227uu20uuuuuuxx uuup%s  c" % language)
    
    # DDC
    setspecs = xmlrecord["header"]["setSpec"]
    for setspec in setspecs:
        if "ddc:" in setspec:
            ddc1 = setspec.replace("ddc:", "")
            break
    else:
        ddc1 = ""        
    
    try:
        subjects = xmlrecord["metadata"]["oai_dc:dc"]["dc:subject"]
    except:
        subjects = []
    for subject in subjects:
        if "ddc" in subject:
            ddc2 = subject.replace("ddc:", "")
            break
    else:
        ddc2 = ""
    
    if len(ddc1) > len(ddc2):
        f082a = ddc1
    else:
        f082a = ddc2
    marcrecord.add("082", a=f082a)

    # Verfasser
    try:
        f100a = xmlrecord["metadata"]["oai_dc:dc"]["dc:creator"]
    except:
        f100a = ""
    marcrecord.add("100", a=f100a)

    # Titel
    f245 = xmlrecord["metadata"]["oai_dc:dc"]["dc:title"]
    if isinstance(f245, list):
        f245a = f245[0]["#text"]
        f245b = f245[1]["#text"]
        marcrecord.add("245", a=f245a, b=f245b)
    else:
        marcrecord.add("245", a=f245["#text"])

    # Erscheinungsvermerk
    f260c = xmlrecord["metadata"]["oai_dc:dc"]["dc:date"]
    publisher = ["b", "Hochschule Mittweida,", "c", f260c]
    marcrecord.add("260", subfields=publisher)

    # Abstract
    try:
        f520a = xmlrecord["metadata"]["oai_dc:dc"]["dc:description"]["#text"]
        marcrecord.add("520", a=f520a)
    except:
        pass

    # Schlagwörter
    try:
        f689a = xmlrecord["metadata"]["oai_dc:dc"]["dc:subject"]
    except:
        f689a = []
    for subject in f689a:
        if "ddc" not in subject:
            if " , " not in subject:                    
                marcrecord.add("689", a=subject)
            else:
                subjects = subject.split(" , ")
                for subject in subjects:
                    marcrecord.add("689", a=subject)

    # Link zu Datensatz und Ressource
    urls = xmlrecord["metadata"]["oai_dc:dc"]["dc:identifier"]
    for url in urls:
        if "docId/" in url:
            marcrecord.add("856", q="text/html", _3="Link zum Datensatz", u=url)
        elif ".pdf" in url or ".PDF" in url:
            marcrecord.add("856", q="text/html", _3="Link zur Ressource", u=url)

    # Medientyp
    marcrecord.add("935", b="cofz")
    marcrecord.add("935", c="hs")

    # Profilierung
    ddc = ddcmatch(f082a)
    if ddc:
        marcrecord.add("650", a="mitddc")
        f980c = ["sid-150-col-monami", "sid-150-col-monamibuch"]
    else:
        f980c = ["sid-150-col-monami"]
    marcrecord.add("980", a=f001, b="150", c=f980c)
    
    outputfile.write(marcrecord.as_marc())

inputfile.close()
outputfile.close()



# coding: utf-8
# pylint: disable=C0301,E1101

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
https://www.ceeol.com/, refs #9398.

Currently, there are two XML formats used, a custom CEEOL XML and a slightly
invalid MARCXML. The last update was manually uploaded to a colorful server
@Telekom.

Config:

[ceeol]

disk-dir = /path/to/disk
updates = /path/to/file.xml, /path/to/another.xml, ...

"""

from __future__ import print_function

import glob
import hashlib
import os
import tempfile

import luigi

from gluish.utils import shellout
from siskin.task import DefaultTask


class CeeolTask(DefaultTask):
    """
    Base task.
    """
    TAG = '53'

    def updateid(self):
        """
        Only use the string of files as hashed identifier. TODO(miku): It would
        be more robust to use the checksums of all input files.
        """
        sha1 = hashlib.sha1()
        sha1.update(self.config.get("ceeol", "updates"))
        return sha1.hexdigest()


class CeeolJournalsUpdates(CeeolTask):
    """
    Create an intermediate schema from zero or more update XML (in MARCXML).
    The output will depend on the set and order of the input files.
    """

    def run(self):
        paths = [p.strip() for p in self.config.get("ceeol", "updates").split(",") if p.strip()]
        _, stopover = tempfile.mkstemp(prefix="siskin-")
        self.logger.debug("found %d updates", len(paths))
        for p in paths:
            shellout("span-import -i ceeol-marcxml {input} >> {output}", input=p, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        filename = "{}.ndj".format(self.updateid())
        return luigi.LocalTarget(path=self.path(filename=filename))


class CeeolJournalsIntermediateSchema(CeeolTask):
    """
    Combine all journals from disk dump and convert them to intermediate
    schema. Add updates. The output name of this task depends on the update id.

    This task depends on a fixed subdirectory, e.g. articles or 1Journal-Articles-XML..
    """

    def requires(self):
        return CeeolJournalsUpdates()

    def run(self):
        diskdir = self.config.get('ceeol', 'disk-dir')
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for path in glob.glob(os.path.join(diskdir, 'articles', 'articles_*xml')):
            shellout("span-import -i ceeol {input} | pigz -c >> {output}", input=path, output=stopover)
        shellout("cat {update} | pigz -c >> {output}", update=self.input().path, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        filename = "{}.ldj.gz".format(self.updateid())
        return luigi.LocalTarget(path=self.path(filename=filename))



#!/usr/bin/env python3
# coding: utf-8

# SID: 109
# Collection: Kunsthochschule für Medien Köln (VK Film)
# Ticket: #8391
# technicalCollectionID: sid-109-col-kunsthochschulekoeln
# Task: khm.py
"""
Review notes:

* get_datafield depends on a name in module scope, brittle
* get_field_* might be a single function
* format = str(format), seems format can be a list, too
* format, large if/else
"""

from __future__ import print_function

import re
import sys

import xmltodict

import marcx
from siskin.utils import marc_build_imprint, xmlstream

formatmap = {
    "Buch": {
        "leader": "cam",
        "007": "tu",
        "935b": "druck"
    },
    u"Mehrbänder": {
        "007": "tu",
        "935b": "druck"
    },
    "DVD": {
        "leader": "ngm",
        "007": "vd",
        "935b": "dvdv",
        "935c": "vide"
    },
    "Blu-ray": {
        "leader": "ngm",
        "007": "vd",
        "935b": "bray",
        "935c": "vide"
    },
    "Videodatei": {
        "leader": "cam",
        "007": "cr",
        "935b": "cofz",
        "935c": "vide"
    },
    "CD": {
        "leader": "  m",
        "007": "c",
        "935b": "cdda"
    },
    "Videokassette": {
        "leader": "cgm",
        "007": "vf",
        "935b": "vika",
        "935c": "vide"
    },
    "Noten": {
        "leader": "nom",
        "007": "zm",
        "935c": "muno"
    },
    "Loseblattsammlung": {
        "leader": "nai",
        "007": "td",
    },
    "Film": {
        "leader": "cam",
        "007": "mu",
        "935b": "sobildtt"
    },
    "Aufsatz": {
        "leader": "naa",
        "007": "tu"
    },
    "Objekt": {
        "leader": "crm",
        "007": "zz",
        "935b": "gegenst"
    },
    "Zeitschrift": {
        "leader": "nas",
        "007": "tu",
        "008": "                     p"
    },
    "Sonstiges": {
        "leader": "npa",
        "007": "tu"
    }
}


def get_datafield(record, tag, code, all=False):
    """
    Return string value for (record, tag, code) or list of values if all is True.
    """
    values = []
    for field in record["ns0:record"]["ns0:datafield"]:
        if field["@tag"] != tag:
            continue
        if isinstance(field["ns0:subfield"], list):
            for subfield in field["ns0:subfield"]:
                if subfield["@code"] != code:
                    continue
                if not all:
                    return subfield["#text"]
                values.append(subfield["#text"])
        else:
            if field["ns0:subfield"]["@code"] == code:
                if not all:
                    return field["ns0:subfield"]["#text"]
                values.append(field["ns0:subfield"]["#text"])
    return values


def get_leader(format="Buch"):
    if format == u"Mehrbänder":
        return "     cam  22       a4500"
    return "     %s  22        4500" % formatmap[format]["leader"]


def get_field_007(format="Buch"):
    return formatmap[format]["007"]


def get_field_008(format="Zeitschrift"):
    if "008" not in formatmap[format]:
        return ""
    return formatmap[format]["008"]


def get_field_935b(format="Buch"):
    if "935b" not in formatmap[format]:
        return ""
    return formatmap[format]["935b"]


def get_field_935c(format="Buch"):
    if "935c" not in formatmap[format]:
        return ""
    return formatmap[format]["935c"]


def remove_brackets(field):
    if isinstance(field, list) and len(field) == 0:
        return ""
    return field.replace("<<", "").replace(">>", "")


inputfilename = "109_input.xml"
outputfilename = "109_output.mrc"

if len(sys.argv) == 3:
    inputfilename, outputfilename = sys.argv[1:]

outputfile = open(outputfilename, "wb")

parent_ids = []
parent_titles = {}

for oldrecord in xmlstream(inputfilename, "record"):

    record = xmltodict.parse(oldrecord)
    parent_id = get_datafield(record, "010", "a")

    if len(parent_id) > 0:
        parent_ids.append(parent_id)

for oldrecord in xmlstream(inputfilename, "record"):

    record = xmltodict.parse(oldrecord)
    id = get_datafield(record, "001", "a")
    title = get_datafield(record, "331", "a")

    if id in parent_ids:
        parent_titles[id] = title

for oldrecord in xmlstream(inputfilename, "record"):

    record = xmltodict.parse(oldrecord)
    marcrecord = marcx.Record(force_utf8=True)
    marcrecord.strict = False

    id = get_datafield(record, "001", "a")
    parent_id = get_datafield(record, "010", "a")
    title = get_datafield(record, "331", "a")

    if "Brockhaus" in title or len(title) == 0:
        continue

    f245a = title
    f245p = ""
    f773w = ""

    if len(parent_id) > 0:
        has_parent_title = parent_titles.get(parent_id, None)
        if has_parent_title:
            f245a = parent_titles[parent_id]
            f245p = title
            f773w = "(DE-576)" + parent_id

    # Format
    format = get_datafield(record, "433", "a")
    format = u'%s' % format
    isbn = get_datafield(record, "540", "a")
    regexp = re.search("S\.\s\d+\s?-\s?\d+", format)

    if id in parent_ids:
        format = u"Mehrbänder"
    elif len(isbn) > 0 and "Videokassette" not in format and "VHS" not in format and "DVD" not in format:
        format = "Buch"
    elif ("S." in format or "Bl." in format or "Ill." in format or " p." in format or "XI" in format or "XV" in format
          or "X," in format or "Bde." in format or ": graph" in format):
        format = "Buch"
    elif "CD" in format:
        format = "CD"
    elif "DVD" in format:
        format = "DVD"
    elif "Blu-ray" in format:
        format = "Blu-ray"
    elif "Videokassette" in format or "VHS" in format or "Min" in format:
        format = "Videokassette"
    elif "Losebl.-Ausg." in format:
        format = "Loseblattsammlung"
    elif regexp:
        format = "Aufsatz"
    elif ("Plakat" in format or "Kassette" in format or "Box" in format or "Karton" in format or "Postkarten" in format
          or "Teile" in format or "USB" in format or "Schachtel" in format or "Schautafel" in format
          or "Medienkombination" in format or "Tafel" in format or "Faltbl" in format or "Schuber" in format):
        format = "Objekt"
    elif id in parent_ids and len(isbn) == 0:  # Zeitschrift 
        continue
    else:
        format = "Buch"

    # Leader
    leader = get_leader(format=format)
    marcrecord.leader = leader

    # Identifier
    f001 = get_datafield(record, "001", "a")
    marcrecord.add("001", data="finc-109-" + f001)

    # 007
    f007 = get_field_007(format=format)
    marcrecord.add("007", data=f007)

    # 008
    f008 = get_field_008(format=format)
    marcrecord.add("008", data=f008)

    # ISBN
    f020a = get_datafield(record, "540", "a")
    marcrecord.add("020", a=f020a)
    f020a = get_datafield(record, "570", "a")
    marcrecord.add("020", a=f020a)

    # Sprache
    f041a = get_datafield(record, "037", "a")
    marcrecord.add("041", a=f041a)

    # 1. Urheber
    f100a = get_datafield(record, "100", "a")
    f100a = remove_brackets(f100a)
    marcrecord.add("100", a=f100a)

    # Haupttitel & Verantwortlichenangabe
    f245a = remove_brackets(f245a)
    f245c = get_datafield(record, "359", "a")
    f245p = remove_brackets(f245p)
    f245 = ["a", f245a, "c", f245c, "p", f245p]
    marcrecord.add("245", subfields=f245)

    # Erscheinungsvermerk
    f260a = get_datafield(record, "410", "a")
    if isinstance(f260a, list):
        f260a = ""
    f260b = get_datafield(record, "412", "a")
    if isinstance(f260b, list):
        f260b = ""
    f260b = remove_brackets(f260b)
    f260c = get_datafield(record, "425", "a")
    if isinstance(f260c, list):
        f260c = ""
    subfields = marc_build_imprint(f260a, f260b, f260c)
    marcrecord.add("260", subfields=subfields)

    # Umfangsangabe
    f300a = get_datafield(record, "433", "a")
    f300a = remove_brackets(f300a)
    f300b = get_datafield(record, "434", "a")
    f300 = ["a", f300a, "b", f300b]
    marcrecord.add("300", subfields=f300)

    f490 = get_datafield(record, "451", "a")
    if len(f490) > 0:
        f490 = f490.split(" ; ")
        if len(f490) == 2:
            f490a = f490[0]
            f490v = f490[1]
        else:
            f490a = f490
            f490v = ""
        marcrecord.add("490", a=f490a, v=f490v)

    for f650a in set(get_datafield(record, "710", "a", all=True)):
        f650a = remove_brackets(f650a)
        marcrecord.add("650", a=f650a)

    for f650a in set(get_datafield(record, "711", "a", all=True)):
        f650a = remove_brackets(f650a)
        marcrecord.add("650", a=f650a)

    # weitere Urheber
    for tag in range(101, 200):
        f700a = get_datafield(record, str(tag), "a")
        f700a = remove_brackets(f700a)
        marcrecord.add("700", a=f700a)

    # weitere Körperschaften
    for tag in range(200, 300):
        f710a = get_datafield(record, str(tag), "a")
        f710a = remove_brackets(f710a)
        marcrecord.add("710", a=f710a)

    # übergeordnetes Werk
    marcrecord.add("773", w=f773w)

    # Links
    f856u = get_datafield(record, "655", "u")
    f8563 = get_datafield(record, "655", "x")
    if len(f8563) == 0:
        f8563 = u"zusätzliche Informationen"
    if "http" in f856u:
        marcrecord.add("856", q="text/html", _3=f8563, u=f856u)

    # Format
    f935b = get_field_935b(format=format)
    f935c = get_field_935c(format=format)
    marcrecord.add("935", b=f935b, c=f935c)

    # Kollektion
    collection = ["a", f001, "b", "109", "c", "sid-109-col-kunsthochschulekoeln"]
    marcrecord.add("980", subfields=collection)

    try:
        outputfile.write(marcrecord.as_marc())
    except UnicodeEncodeError as exc:
        print("%s: %s" % (marcrecord["001"], exc), file=sys.stderr)

outputfile.close()



#!/usr/bin/env python
# coding: utf-8

import io
import re
import sys
# https://stackoverflow.com/a/40846742/89391
import warnings

import pandas

import marcx
import pymarc

warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")

inputfilename = "160_input.csv"
outputfilename = "160_output.mrc"

if len(sys.argv) == 3:
    inputfilename, outputfilename = sys.argv[1:]

outputfile = io.open(outputfilename, "wb")
writer = pymarc.MARCWriter(outputfile)

csv_records = pandas.read_csv(inputfilename, encoding="latin-1", sep=";")
for csv_record in csv_records.iterrows():

    csv_record = csv_record[1]
    marc_record = marcx.Record(force_utf8=True)

    marc_record.leader = "     nam  22        4500"

    f001 = "finc-160-" + str(csv_record["001"])
    marc_record.add("001", data=f001)

    # Zugangsformat
    marc_record.add("007", data="tu")
    
    # Sprache
    marc_record.add("041", a=csv_record["041a"])
    
    # Notation
    marc_record.add("084", a="ZX 3900", _2="rvk")
   
    # 1. Urheber
    marc_record.add("100", a=csv_record["100a"])
    
    # Titel
    marc_record.add("245", a=csv_record["245a"])

    # Erscheinungsvermerk
    publisher = csv_record["502a"]
    match1 = re.search("(.*?),\s(.*?),\s(.*?),\s([\[\d].*)", publisher)
    match2 = re.search("(.*?),\s(.*?),\s(.*?),\s(.*?),\s([\[\d].*)", publisher)
    if match1:
        f260a = match1.group(1)
        f260b = match1.group(2)
    elif match2:
        f260a = match2.group(1)
        f260b1 = match2.group(2)
        f260b2 = match2.group(3)
        f260b = f260b1 + ", " + f260b2
    else:
        f260a = ""
        f260b = ""
    publisher = ["a", f260a, "b", f260b, "c", csv_record["260c"]]
    marc_record.add("260", subfields=publisher)

    # Umfang
    marc_record.add("300", a=csv_record["300a"])
    
    # Fußnote
    marc_record.add("500", a=u"Signatur: " + csv_record["Signatur\ngesamt"])

    # Hochschulvermerk
    university = csv_record["502a"]
    match1 = re.search("(.*?),\s(.*?),\s(.*?),\s([\[\d].*)", university)
    match2 = re.search("(.*?),\s(.*?),\s(.*?),\s(.*?),\s([\[\d].*)", university)
    if match1 or match2:
        marc_record.add("502", a=university)

    # weitere Urheber
    for field in csv_record.keys():
        if not field.startswith("700a"):
            continue
        if pandas.isnull(csv_record[field]):
            continue
        marc_record.add("700", a=csv_record[field])

    # Medienform
    marc_record.add("935", b=csv_record["935b"], c=csv_record["935c"])

    # Kollektion
    marc_record.add("980", a=str(csv_record["001"]), b="160", c="sid-160-col-diplspowi")

    writer.write(marc_record)

writer.close()



# coding: utf-8
# pylint: disable=C0301,E1101,C0330,C0111

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
Electronic Resource Management System Based on Linked Data Technologies.

https://amsl.technology

> Managing electronic resources has become a distinctive and important task for
libraries in recent years. The diversity of resources, changing licensing
policies and new business models of the publishers, consortial acquisition and
modern web scale discovery technologies have turned the market place of
scientific information into a complex and multidimensional construct.

These tasks wrap API responses, allow access to holding files, and create a
filterconfig which can be use fed into
[span-tag](https://github.com/miku/span/tree/master/cmd/span-tag).

Config:

[amsl]

base = https://x.y.z
uri-download-prefix = https://x.y.z/OntoWiki/files/get?setResource=
write-url = https://x.y.z/OntoWiki/x/y # For time-stamping.

"""

from __future__ import print_function

import collections
import datetime
import gzip
import io
import itertools
import json
import operator
import tempfile
import zipfile

import luigi

from gluish.format import TSV, Gzip
from gluish.utils import shellout
from siskin.task import DefaultTask
from siskin.utils import SetEncoder, dictcheck


class AMSLTask(DefaultTask):
    """
    Base class for AMSL related tasks.
    """
    TAG = 'amsl'


class AMSLServiceDeprecated(AMSLTask):
    """
    Defunkt task via #14415 as of 2018-12-12. Will be remove soon.

    Retrieve AMSL API response. Outbound: holdingsfiles, contentfiles, metadata_usage.
    2018-12-12: discovery API EOL, XXX: adjust, refs #14415.

    Example output (discovery):

        [
            {
                "shardLabel": "SLUB-dbod",
                "sourceID": "64",
                "megaCollection": "Perinorm – Datenbank für Normen und technische Regeln",
                "productISIL": null,
                "externalLinkToContentFile": null,
                "contentFileLabel": null,
                "contentFileURI": null,
                "linkToContentFile": null,
                "ISIL": "DE-105",
                "evaluateHoldingsFileForLibrary": "no",
                "holdingsFileLabel": null,
                "holdingsFileURI": null,
                "linkToHoldingsFile": null
            },
            {
                "shardLabel": "SLUB-dbod",
                "sourceID": "64",
                "megaCollection": "Perinorm – Datenbank für Normen und technische Regeln",
                "productISIL": null,
                "externalLinkToContentFile": null,
                "contentFileLabel": null,
                "contentFileURI": null,
                "linkToContentFile": null,
                "ISIL": "DE-14",
                "evaluateHoldingsFileForLibrary": "no",
                "holdingsFileLabel": null,
                "holdingsFileURI": null,
                "linkToHoldingsFile": null
            },
        ...

    """
    date = luigi.DateParameter(default=datetime.date.today())
    name = luigi.Parameter(default='outboundservices:discovery',
                           description='discovery, holdingsfiles, contentfiles, metadata_usage, freeContent')

    def run(self):
        parts = self.name.split(':')
        if not len(parts) == 2:
            raise RuntimeError('realm:name expected, e.g. outboundservices:discovery')
        realm, name = parts

        link = '%s/%s/list?do=%s' % (self.config.get('amsl', 'base').rstrip('/'), realm, name)
        output = shellout("""curl --fail "{link}" | pigz -c > {output} """, link=link)

        # Check for valid JSON before, simplifies debugging.
        with gzip.open(output, 'rb') as handle:
            try:
                _ = json.load(handle)
            except ValueError as err:
                self.logger.warning("AMSL API did not return valid JSON")
                raise

        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True, ext='json.gz'), format=Gzip)


class AMSLService(AMSLTask):
    """
    Approximation of discovery response until better approach is implemented. Requires span 0.1.273 or later.
    """
    date = luigi.DateParameter(default=datetime.date.today())
    name = luigi.Parameter(default='outboundservices:discovery', description='ignored, kept for compatibility')

    def requires(self):
        if self.name != 'outboundservices:discovery':
            return AMSLServiceDeprecated(date=self.date, name=self.name)
        return []

    def run(self):
        if self.name == 'outboundservices:discovery':
            output = shellout("span-amsl-discovery -live {live} | gzip -c > {output}",
                              live=self.config.get('amsl', 'base'))
        else:
            output = self.input().path

        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True, ext='json.gz'), format=Gzip)


class AMSLCollections(AMSLTask):
    """
    Report all collections, that appear in AMSL.
    """
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return AMSLService(date=self.date, name='outboundservices:discovery')

    def run(self):
        output = shellout(""" unpigz -c {input} | jq -rc '.[]|.megaCollection' | sort -u > {output}""",
                          input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class AMSLCollectionsShardFilter(AMSLTask):
    """
    A per-shard list of collection entries. One record per line.

        {
          "evaluateHoldingsFileForLibrary": "no",
          "holdingsFileLabel": null,
          "megaCollection": "DOAJ Directory of Open Access Journals",
          "shardLabel": "UBL-ai",
          "contentFileURI": null,
          "sourceID": "28",
          "linkToHoldingsFile": null,
          "holdingsFileURI": null,
          "productISIL": null,
          "linkToContentFile": null,
          "contentFileLabel": null,
          "externalLinkToContentFile": null,
          "ISIL": "DE-14"
        }
        ....

    Shard distribution as of January 2017:

        $ taskcat AMSLService | jq -rc '.[] | .shardLabel' | sort | uniq -c | sort -nr
        53493 UBL-ai
         1121 UBL-main
          245 SLUB-dswarm
           19 SLUB-dbod

    """
    date = luigi.DateParameter(default=datetime.date.today())
    shard = luigi.Parameter(default='UBL-ai', description='only collect items for this shard')

    def requires(self):
        return AMSLService(date=self.date, name='outboundservices:discovery')

    def run(self):
        with self.input().open() as handle:
            doc = json.load(handle)

        with self.output().open('w') as output:
            for item in doc:
                if not item['shardLabel'] == self.shard:
                    continue
                output.write(json.dumps(item) + "\n")

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class AMSLCollectionsISILList(AMSLTask):
    """
    A per-shard list of ISILs for which AMSL has some information.

        DE-105
        DE-14
        DE-15
        DE-15-FID
        DE-1972
        DE-540
        ...
    """
    date = luigi.DateParameter(default=datetime.date.today())
    shard = luigi.Parameter(default='UBL-ai', description='only collect items for this shard')

    def requires(self):
        return AMSLService(date=self.date, name='outboundservices:discovery')

    def run(self):
        with self.input().open() as handle:
            doc = json.load(handle)

        isils = set()

        for item in doc:
            if not item['shardLabel'] == self.shard:
                continue
            isils.add(item['ISIL'])

        if len(isils) == 0:
            raise RuntimeError('no isils found: maybe mispelled shard name?')

        with self.output().open('w') as output:
            for isil in sorted(isils):
                output.write_tsv(isil)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class AMSLCollectionsISIL(AMSLTask):
    """
    Per ISIL list of collections.

        {
          "48": [
            "Genios (Recht)",
            "Genios (Sozialwissenschaften)",
            "Genios (Psychologie)",
            "Genios (Fachzeitschriften)",
            "Genios (Wirtschaftswissenschaften)"
          ],
          "49": [
            "Helminthological Society (CrossRef)",
            "International Association of Physical Chemists (IAPC) (CrossRef)",
            ...

    Examples (Jan 2017):

        $ taskcat AMSLCollectionsISIL --isil DE-Mit1 --shard SLUB-dbod | jq .
        {
            "64": [
                "Perinorm – Datenbank für Normen und technische Regeln"
            ]
        }

        $ taskcat AMSLCollectionsISIL --isil DE-14 --shard SLUB-dswarm | jq .
        {
            "83": [
                "SLUB/Mediathek"
            ],
            "69": [
                "Wiley ebooks"
            ],
            "105": [
                "Springer Journals"
            ],
            "94": [
                "Blackwell Publishing Journal Backfiles 1879-2005",
                "China Academic Journals (CAJ) Archiv",
                "Wiley InterScience Backfile Collections 1832-2005",
                "Torrossa / Periodici",
                "Cambridge Journals Digital Archive",
                "Elsevier Journal Backfiles on ScienceDirect",
                "Periodicals Archive Online",
                "Emerald Fulltext Archive Database",
                "Springer Online Journal Archives"
            ],
            "67": [
                "SLUB/Deutsche Fotothek"
            ]
        }

        $ taskcat AMSLCollectionsISIL --isil DE-15 --shard SLUB-dswarm | jq .
        {
            "68": [
                "OLC SSG Medien- / Kommunikationswissenschaft",
                "OLC SSG Film / Theater"
            ],
            "94": [
                "Blackwell Publishing Journal Backfiles 1879-2005",
                "China Academic Journals (CAJ) Archiv",
                "Wiley InterScience Backfile Collections 1832-2005",
                "Torrossa / Periodici",
                "Cambridge Journals Digital Archive",
                "Elsevier Journal Backfiles on ScienceDirect",
                "Periodicals Archive Online",
                "Emerald Fulltext Archive Database",
                "Springer Online Journal Archives"
            ],
            "105": [
                "Springer Journals"
            ]
        }
    """
    date = luigi.DateParameter(default=datetime.date.today())
    isil = luigi.Parameter(default='DE-15', description='ISIL, case sensitive')
    shard = luigi.Parameter(default='UBL-ai', description='only collect items for this shard')

    def requires(self):
        return AMSLService(date=self.date, name='outboundservices:discovery')

    def run(self):
        with self.input().open() as handle:
            doc = json.load(handle)

        unique_shards = set()

        scmap = collections.defaultdict(set)
        for item in doc:
            unique_shards.add(item['shardLabel'])
            if not item['shardLabel'] == self.shard:
                continue
            if not item['ISIL'] == self.isil:
                continue
            scmap[item['sourceID']].add(item['megaCollection'].strip())
            if not scmap:
                raise RuntimeError('no collections found for ISIL: %s' % self.isil)

        if not scmap and self.shard not in unique_shards:
            self.logger.warn('available shards: %s', list(unique_shards))

        with self.output().open('w') as output:
            output.write(json.dumps(scmap, cls=SetEncoder) + "\n")

    def output(self):
        return luigi.LocalTarget(path=self.path())


class AMSLHoldingsFile(AMSLTask):
    """
    Access AMSL files/get?setResource= facilities.

    The output is probably zipped (will be decompressed on the fly).

    One ISIL can have multiple files (they will be concatenated).

    Output should be in standard KBART format, given the uploaded files in AMSL are KBART.
    """
    isil = luigi.Parameter(default='DE-15', description='ISIL, case sensitive')
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return AMSLService(date=self.date, name='outboundservices:holdingsfiles')

    def run(self):
        with self.input().open() as handle:
            holdings = json.load(handle)

        _, stopover = tempfile.mkstemp(prefix='siskin-')

        # The property containing the URI of the holding file, maybe.
        urikey = 'DokumentURI'

        for holding in holdings:
            if holding["ISIL"] == self.isil:

                if urikey not in holding:
                    raise RuntimeError('possible AMSL API change, expected: %s, available keys: %s' %
                                       (urikey, list(holding.keys())))

                # refs. #7142
                if 'kbart' not in holding[urikey].lower():
                    self.logger.debug("skipping non-KBART holding URI: %s", holding[urikey])
                    continue

                link = "%s%s" % (self.config.get('amsl', 'uri-download-prefix'), holding[urikey])
                downloaded = shellout("curl --fail {link} > {output} ", link=link)
                try:
                    _ = zipfile.ZipFile(downloaded)
                    shellout("unzip -p {input} >> {output}", input=downloaded, output=stopover)
                except zipfile.BadZipfile:
                    # Probably not a zip.
                    shellout("cat {input} >> {output}", input=downloaded, output=stopover)

        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())


class AMSLOpenAccessISSNList(AMSLTask):
    """
    List of ISSN, which are open access.

    For now, extract these ISSN from a special holding file, called
    KBART_FREEJOURNALS via AMSL.

    Example (Jan 2017):

        $ taskcat AMSLOpenAccessISSNList | head -10
        0001-0944
        0001-1843
        0001-186X
        0001-1983
        0001-2114
        0001-2211
        0001-267X
        0001-3714
        0001-3757
        0001-3765

    As of October 2017, this list includes: https://pub.uni-bielefeld.de/download/2913654/2913655.
    As of December 2018, there are 72692 ISSN listed.
    """

    date = luigi.DateParameter(default=datetime.date.today())

    def run(self):
        """
        Download, maybe unzip, grab column 2 and 3, keep ISSN, sort -u.
        """
        key = "http://amsl.technology/discovery/metadata-usage/Dokument/KBART_FREEJOURNALS"
        link = "%s%s" % (self.config.get('amsl', 'uri-download-prefix'), key)

        downloaded = shellout("curl --fail {link} > {output} ", link=link)
        _, stopover = tempfile.mkstemp(prefix='siskin-')

        try:
            _ = zipfile.ZipFile(downloaded)
            output = shellout("unzip -p {input} >> {output}", input=downloaded)
        except zipfile.BadZipfile:
            # At least the file is not a zip.
            output = shellout("cat {input} >> {output}", input=downloaded)

        shellout("cut -f 2 {input} | grep -oE '[0-9]{{4,4}}-[xX0-9]{{4,4}}' >> {output}", input=output, output=stopover)
        shellout("cut -f 3 {input} | grep -oE '[0-9]{{4,4}}-[xX0-9]{{4,4}}' >> {output}", input=output, output=stopover)

        # Include OA list, refs #11579, maybe cache this?
        shellout("""curl -s https://pub.uni-bielefeld.de/download/2913654/2913655 | cut -d, -f1,2 | tr -d '"' |
                    grep -E '[0-9]{{4,4}}-[0-9]{{3,3}}[0-9xX]' | tr ',' '\n' >> {output}""",
                 output=stopover,
                 preserve_whitespace=True)

        output = shellout("sort -u {input} > {output}", input=stopover)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())


class AMSLGoldListKBART(AMSLTask):
    """
    Convert Bielefeld Gold List to KBART (for manual uploads in AMSL). As of
    December 2018, there are 36706 ISSN in this list.
    """

    date = luigi.DateParameter(default=datetime.date.today())

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        shellout(""" echo "online_identifier" > {output}""", output=stopover)
        # Include OA list, refs #11579.
        shellout("""curl -s https://pub.uni-bielefeld.de/download/2913654/2913655 |
                    cut -d, -f1,2 |
                    tr -d '"' |
                    grep -E '[0-9]{{4,4}}-[0-9]{{3,3}}[0-9xX]' |
                    tr ',' '\n' |
                    sort -u |
                    grep -v ^$ >> {output}""",
                 output=stopover,
                 preserve_whitespace=True)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())


class AMSLFreeContent(AMSLTask):
    """
    Free content. Revelant for OA flags.
    """
    date = luigi.DateParameter(default=datetime.date.today())

    def run(self):
        output = shellout("curl -s '{base}/inhouseservices/list?do=freeContent' | jq -rc . > {output}",
                          base=self.config.get('amsl', 'base'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())


class AMSLOpenAccessKBART(AMSLTask):
    """
    Create a KBART file that contains open access and freely available journals only.

    Used in conjunction with [span-oa-filter](https://git.io/vdB29).
    """

    date = luigi.DateParameter(default=datetime.date.today())

    def run(self):
        """
        Download, maybe unzip, combine with Gold List.
        """
        key = "http://amsl.technology/discovery/metadata-usage/Dokument/KBART_FREEJOURNALS"
        link = "%s%s" % (self.config.get('amsl', 'uri-download-prefix'), key)

        downloaded = shellout("curl --fail {link} > {output} ", link=link)
        _, stopover = tempfile.mkstemp(prefix='siskin-')

        try:
            _ = zipfile.ZipFile(downloaded)
            output = shellout("unzip -p {input} >> {output}", input=downloaded)
        except zipfile.BadZipfile:
            # At least the file is not a zip.
            output = shellout("cat {input} >> {output}", input=downloaded)

        # Include OA list, refs #11579.
        shellout("""curl -s https://pub.uni-bielefeld.de/download/2913654/2913655 | cut -d, -f1,2 | tr -d '"' |
                    grep -E '[0-9]{{4,4}}-[0-9]{{3,3}}[0-9xX]' | tr ',' '\n' |
                    awk '{{ print "\t\t"$0 }}' >> {output}""",
                 output=output,
                 preserve_whitespace=True)

        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())


class AMSLWisoPackages(AMSLTask):
    """
    Collect WISO packages.

    XXX(miku): Throw this away.
    """
    date = luigi.DateParameter(default=datetime.date.today())
    encoding = luigi.Parameter(default='latin-1', description='wiso journal id csv file encoding', significant=False)

    def requires(self):
        return AMSLService(date=self.date)

    def hardcoded_list_of_wiso_journal_identifiers(self):
        """
        Refs. #10707, Att. #4444.
        """
        ids = set()
        filename = self.assets('wiso/645896059854847ce4ccd1416e11ba372e45bfd6.csv')
        with io.open(filename, encoding=self.encoding) as handle:
            for i, line in enumerate(handle):
                if i == 0:
                    continue
                line = line.strip()
                if not line:
                    continue
                fields = line.split(';')
                if len(fields) < 12:
                    raise ValueError('expected ; separated KBART-ish file with journal identifier at column 11')
                id = fields[11].strip()
                if not id:
                    continue
                ids.add(id)

        return sorted(ids)

    def resolve_ubl_profile(self, colls):
        """
        Given a list of collection names, replace the complete list with
        hardcoded WISO journal identifiers from #10707/4444.
        """
        if 'wiso UB Leipzig Profil' in colls:
            return self.hardcoded_list_of_wiso_journal_identifiers()

        return colls

    def run(self):
        with self.input().open() as handle:
            doc = json.loads(handle.read())

        isilpkg = collections.defaultdict(lambda: collections.defaultdict(set))

        for item in doc:
            isil, sid = item.get('ISIL'), item.get('sourceID')
            mega_collection = item.get('megaCollection')
            lthf = item.get('linkToHoldingsFile')
            if sid != "48":
                continue
            isilpkg[isil][lthf].add(mega_collection)

        filterconfig = collections.defaultdict(dict)
        fzs_package_name = 'Genios (Fachzeitschriften)'

        for isil, blob in list(isilpkg.items()):
            include_fzs = False
            for _, colls in list(blob.items()):
                if fzs_package_name in colls:
                    include_fzs = True

            filters = []

            if include_fzs and isil != 'DE-15-FID':
                packages = set(itertools.chain(*[c for _, c in list(blob.items())]))
                packages = self.resolve_ubl_profile(packages)
                filters.append({
                    'and': [{
                        'source': ['48']
                    }, {
                        'package': [fzs_package_name]
                    }, {
                        'package': [name for name in packages if name != fzs_package_name]
                    }]
                })

            for lthf, colls in list(blob.items()):
                if lthf is None or lthf == 'null':
                    continue
                if isil == 'DE-15-FID':
                    colls = self.resolve_ubl_profile(colls)
                    filter = {
                        'and': [
                            {
                                'source': ['48']
                            },
                            {
                                'holdings': {
                                    'urls': [lthf]
                                }
                            },
                            {
                                'package': [c for c in colls if c != fzs_package_name]
                            },
                        ]
                    }
                else:
                    filter = {
                        'and': [{
                            'source': ['48']
                        }, {
                            'holdings': {
                                'urls': [lthf]
                            }
                        }, {
                            'package': [c for c in colls if c != fzs_package_name]
                        }, {
                            'not': {
                                'package': [fzs_package_name]
                            }
                        }]
                    }
                filters.append(filter)
            filterconfig[isil] = {'or': filters}

        with self.output().open('w') as output:
            json.dump(filterconfig, output, cls=SetEncoder)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='json'))


class AMSLFilterConfigFreeze(AMSLTask):
    """
    Create a frozen file. File will contain the filterconfig plus content of all URLs.
    """
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return AMSLFilterConfig(date=self.date)

    def run(self):
        output = shellout("span-freeze -o {output} < {input}", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='zip'))


class AMSLFilterConfig(AMSLTask):
    """
    Turn AMSL API to a span(1) filter configuration.

    Cases (Spring 2017):

    61084 Case 3 (ISIL, SID, Collection, Holding File)
      805 Case 2 (ISIL, SID, Collection, Product ISIL)
      380 Case 1 (ISIL, SID, Collection)
       38 Case 4 (ISIL, SID, Collection, External Content File)
       31 Case 6 (ISIL, SID, Collection, External Content File, Holding File)
        1 Case 7 (ISIL, SID, Collection, Internal Content File, Holding File)
        1 Case 5 (ISIL, SID, Collection, Internal Content File)

    Process:

                AMSL Discovery API
                      |
                      v
                AMSLFilterConfig
                      |
                      v
        $ span-tag -c config.json < input.is > output.is

    Notes:

    This task turns an AMSL discovery API response into a filterconfig[1], which
    span(1) can understand.

    AMSL API might not specify everything we need to know, so this task shall
    be the only place, where workarounds happen.

    Whil span-tag is fast, it is not fast enough to iterate over a disjuction
    of 60K items for each of the 100M documents fast enough, which - if we
    could - would simplify the implementation of this task.

    The main speed improvement comes from using lists of collection names
    instead of having each collection processed separately - which is how it
    works conceptually: Each collection could use a separate KBART file (or use
    none at all).

    We ignore collection names, if (external) content files are used. These
    content files are usually there, because we cannot infer the collection
    name from the data alone.

    Performance data point: 22 ISIL each with between 1 and 26 alternatives for
    attachment, each alternative consisting of around three filters. Around 30
    holding or content files each with between 10 and 50000 entries referenced
    about 200 times in total: around 20k records/s.

    Case table (Feb 2019), X/-/o, yes, no, maybe.

    SID COLL ISIL LTHF LTCF ELTCF PI TCID
    -------------------------------------
    X   X    X    -    -    -     -   o
    X   X    X    X    -    -     -   o
    X   X    X    X    X    -     -   o
    X   X    X    X    -    X     -   o
    X   X    X    X    -    -     X   o
    X   X    X    -    X    -     -   o
    X   X    X    -    -    X     -   o
    X   X    X    -    -    -     X   o

    For a transition period, we extend the collection list from AMSL with
    canoncical names via DOI prefix, see #13587. The mapping is fixed and need
    to be updated manually.

    ----

    [1] https://git.io/vQohE
    """

    date = luigi.DateParameter(default=datetime.date.today())

    def extend_collections(self, colls):
        """
        Given a list of collection names, extend the list by adding the canonicals names as well, refs #13587.
        """
        if not hasattr(self, '_name_to_canonical'):
            with open(self.assets("amsl/13587.json")) as handle:
                self._name_to_canonical = json.load(handle)

        result = set()
        for c in colls:
            result.add(c)
            if c in self._name_to_canonical:
                result.add(self._name_to_canonical[c])

        self.logger.debug("extended collection list from %d to %d items (%d mappings)" %
                          (len(colls), len(result), len(self._name_to_canonical)))
        return list(result)

    def requires(self):
        return {
            'amsl': AMSLService(date=self.date),
            'wiso': AMSLWisoPackages(date=self.date),
        }

    def run(self):
        with self.input().get('amsl').open() as handle:
            doc = json.loads(handle.read())

        # Case: ISIL, SID, collection.
        isilsidcollections = collections.defaultdict(lambda: collections.defaultdict(set))

        # Case: ISIL, SID, collection, link.
        isilsidlinkcollections = collections.defaultdict(lambda: collections.defaultdict(lambda: collections.
                                                                                         defaultdict(set)))

        # Ready-made filters per ISIL. Some filters can be added on-the-fly
        # because there aren't many occurences.
        isilfilters = collections.defaultdict(list)

        for item in doc:
            isil, sid, mega_collection, technicalCollectionID = operator.itemgetter('ISIL', 'sourceID',
                                                                                    'megaCollection',
                                                                                    'technicalCollectionID')(item)

            if sid == '48':  # Handled elsewhere.
                continue

            # refs #10495, a subject filter for a few hard-coded ISIL.
            if sid == '34' and isil in ('DE-L152', 'DE-1156', 'DE-1972', 'DE-Kn38'):
                isilfilters[isil].append({"and": [
                    {
                        "source": ["34"],
                    },
                    {
                        "subject": [
                            "Music",
                            "Music education",
                        ]
                    },
                ]})
                continue

            # refs #10495, maybe use a TSV with custom column name to use a subject list?
            if sid == '34' and isil == 'DE-15-FID':
                isilfilters[isil].append({
                    "and": [
                        {
                            "source": ["34"],
                        },
                        {
                            "subject": [
                                "Film studies",
                                "Information science",
                                "Mass communication",
                            ]
                        },
                    ]
                })
                continue

            # SID COLL ISIL LTHF LTCF ELTCF PI TCID
            # -------------------------------------
            # X   X    X    -    -    -     -   o
            if dictcheck(item,
                         contains=['sourceID', 'megaCollection', 'ISIL'],
                         absent=['linkToHoldingsFile', 'linkToContentFile', 'externalLinkToContentFile', 'productISIL'],
                         ignore=['technicalCollectionID']):

                isilsidcollections[isil][sid].add(mega_collection)
                if technicalCollectionID:
                    isilsidcollections[isil][sid].add(technicalCollectionID)

            # SID COLL ISIL LTHF LTCF ELTCF PI TCID
            # -------------------------------------
            # X   X    X    -    -    -     X   o
            elif dictcheck(item,
                           contains=['sourceID', 'megaCollection', 'ISIL', 'productISIL'],
                           absent=['linkToHoldingsFile', 'linkToContentFile', 'externalLinkToContentFile'],
                           ignore=['technicalCollectionID']):

                isilsidcollections[isil][sid].add(mega_collection)
                self.logger.debug("productISIL given, but ignored: %s, %s, %s", isil, sid, item['productISIL'])
                if technicalCollectionID:
                    isilsidcollections[isil][sid].add(technicalCollectionID)

            # SID COLL ISIL LTHF LTCF ELTCF PI TCID
            # -------------------------------------
            # X   X    X    X    -    -     X   o
            elif dictcheck(item,
                           contains=['sourceID', 'megaCollection', 'ISIL', 'linkToHoldingsFile', 'productISIL'],
                           absent=['linkToContentFile', 'externalLinkToContentFile'],
                           ignore=['technicalCollectionID']):

                self.logger.debug("productISIL is set, but we do not have a filter for it yet: %s, %s, %s", isil, sid,
                                  mega_collection)

                if item.get('evaluateHoldingsFileForLibrary') == "yes":
                    isilsidlinkcollections[isil][sid][item['linkToHoldingsFile']].add(mega_collection)
                    if technicalCollectionID:
                        isilsidlinkcollections[isil][sid][item['linkToHoldingsFile']].add(technicalCollectionID)
                else:
                    self.logger.warning("evaluateHoldingsFileForLibrary=no plus link: skipping %s", item)

            # SID COLL ISIL LTHF LTCF ELTCF PI TCID
            # -------------------------------------
            # X   X    X    X    -    -     -   o
            elif dictcheck(item,
                           contains=['sourceID', 'megaCollection', 'ISIL', 'linkToHoldingsFile'],
                           absent=['linkToContentFile', 'externalLinkToContentFile', 'productISIL'],
                           ignore=['technicalCollectionID']):

                if item.get('evaluateHoldingsFileForLibrary') == "yes":
                    isilsidlinkcollections[isil][sid][item['linkToHoldingsFile']].add(mega_collection)
                    if technicalCollectionID:
                        isilsidlinkcollections[isil][sid][item['linkToHoldingsFile']].add(technicalCollectionID)
                else:
                    self.logger.warning("evaluateHoldingsFileForLibrary=no plus link: skipping %s", item)

            # SID COLL ISIL LTHF LTCF ELTCF PI TCID
            # -------------------------------------
            # X   X    X    -    -    X     -   o
            elif dictcheck(item,
                           contains=['sourceID', 'megaCollection', 'ISIL', 'externalLinkToContentFile'],
                           absent=['linkToHoldingsFile', 'linkToContentFile', 'productISIL'],
                           ignore=['technicalCollectionID']):

                isilfilters[isil].append(
                    {"and": [
                        {
                            "source": [sid]
                        },
                        {
                            "holdings": {
                                "urls": [item["externalLinkToContentFile"]]
                            }
                        },
                    ]})

            # SID COLL ISIL LTHF LTCF ELTCF PI TCID
            # -------------------------------------
            # X   X    X    -    X    -     -   o
            elif dictcheck(item,
                           contains=['sourceID', 'megaCollection', 'ISIL', 'linkToContentFile'],
                           absent=['linkToHoldingsFile', 'externalLinkToContentFile', 'productISIL'],
                           ignore=['technicalCollectionID']):

                isilfilters[isil].append(
                    {"and": [
                        {
                            "source": [sid]
                        },
                        {
                            "holdings": {
                                "urls": [item["linkToContentFile"]]
                            }
                        },
                    ]})

            # SID COLL ISIL LTHF LTCF ELTCF PI TCID
            # -------------------------------------
            # X   X    X    X    -    X     -   o
            elif dictcheck(
                    item,
                    contains=['sourceID', 'megaCollection', 'ISIL', 'linkToHoldingsFile', 'externalLinkToContentFile'],
                    absent=['linkToContentFile', 'productISIL'],
                    ignore=['technicalCollectionID']):

                if item.get('evaluateHoldingsFileForLibrary') == "yes":
                    isilfilters[isil].append({
                        "and": [
                            {
                                "source": [sid]
                            },
                            {
                                "holdings": {
                                    "urls": [item["externalLinkToContentFile"]]
                                }
                            },
                            {
                                "holdings": {
                                    "urls": [item["linkToHoldingsFile"]]
                                }
                            },
                        ]
                    })
                else:
                    self.logger.warning("evaluateHoldingsFileForLibrary=no plus link: skipping %s", item)

            # SID COLL ISIL LTHF LTCF ELTCF PI TCID
            # -------------------------------------
            # X   X    X    X    X    -     -   o
            elif dictcheck(item,
                           contains=['sourceID', 'megaCollection', 'ISIL', 'linkToHoldingsFile', 'linkToContentFile'],
                           absent=['externalLinkToContentFile', 'productISIL'],
                           ignore=['technicalCollectionID']):

                if item.get('evaluateHoldingsFileForLibrary') == "yes":
                    isilfilters[isil].append({
                        "and": [
                            {
                                "source": [sid]
                            },
                            {
                                "holdings": {
                                    "urls": [item["linkToContentFile"]]
                                }
                            },
                            {
                                "holdings": {
                                    "urls": [item["linkToHoldingsFile"]]
                                }
                            },
                        ]
                    })
                else:
                    self.logger.warning("evaluateHoldingsFileForLibrary=no plus link: skipping %s", item)

            # SID COLL ISIL LTHF LTCF ELTCF PI TCID
            # -------------------------------------
            # ?   ?    ?    ?    ?    ?     ?   o
            else:
                raise RuntimeError("unhandled combination of sid, collection and other parameters: %s", item)

        # A second pass.
        for isil, blob in list(isilsidcollections.items()):
            for sid, colls in list(blob.items()):
                isilfilters[isil].append(
                    {"and": [
                        {
                            "source": [sid]
                        },
                        {
                            "collection": sorted(self.extend_collections(colls))
                        },
                    ]})

        # A second pass.
        for isil, blob in list(isilsidlinkcollections.items()):
            for sid, spec in list(blob.items()):
                for link, colls in list(spec.items()):
                    isilfilters[isil].append({
                        "and": [
                            {
                                "source": [sid]
                            },
                            {
                                "collection": sorted(self.extend_collections(colls))
                            },
                            {
                                "holdings": {
                                    "urls": [link]
                                }
                            },
                        ]
                    })

        # Final assembly.
        filterconfig = collections.defaultdict(dict)
        for isil, filters in list(isilfilters.items()):
            if len(filters) == 0:
                continue
            if len(filters) == 1:
                filterconfig[isil] = filters[0]
                continue
            filterconfig[isil] = {"or": filters}

        # Include WISO.
        with self.input().get('wiso').open() as handle:
            wisoconf = json.load(handle)
            for isil, tree in list(wisoconf.items()):
                for filter in tree.get('or', []):
                    filterconfig[isil]['or'].append(filter)

        # XXX: Adjust a few items for DE-14, cf. 2018-06-11, namely, add links
        # to external holding files, which are not included into the AMSL
        # discovery API response, refs #13378.
        fix_url = 'https://dbod.de/SLUB-EZB-KBART.zip'
        for term in filterconfig["DE-14"]["or"]:
            for t in term["and"]:
                if (not 'holdings' in t) and (not 'urls' in t.get('holdings', [])):
                    continue
                if fix_url in t['holdings']['urls']:
                    continue
                t['holdings']['urls'].append("https://dbod.de/SLUB-EZB-KBART.zip")

        with self.output().open('w') as output:
            json.dump(filterconfig, output, cls=SetEncoder)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='json'))


class AMSLCollectionList(AMSLTask):
    """
    TODO: implement.

    Find all collection names used in AMSL and their canonical name, via
    members API, prefix and names.
    """

    def run(self):
        raise NotImplementedError()



# coding: utf-8
# pylint: disable=C0301,E1101

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
DawsonEra, #9773, SID124.

Experimental: zip, marcxml, missing ns, illegal characters.

Configuration
-------------

[dawson]

download-url = http://exa.com/store/123

"""

import luigi

from gluish.utils import shellout
from siskin.task import DefaultTask


class DawsonTask(DefaultTask):
    """
    Base class for Dawson.
    """
    TAG = '124'


class DawsonDownload(DawsonTask):
    """
    Download.
    """

    def run(self):
        output = shellout("curl --fail -sL {url} > {output}", url=self.config.get('dawson', 'download-url'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='zip'))


class DawsonFixAndCombine(DawsonTask):
    """
    Remove namespace, combine all files into a single file.

    > Exception in thread "main" org.culturegraph.mf.exceptions.MetafactureException:
    org.xml.sax.SAXParseException; lineNumber: 540690; columnNumber: 68; An invalid
    XML character (Unicode: 0x1a) was found in the element content of the document.
    """

    def requires(self):
        return DawsonDownload()

    def run(self):
        output = shellout(r"""
            echo '<?xml version="1.0" encoding="UTF-8"?>
            <collection xmlns="http://www.loc.gov/MARC21/slim">
            ' >> {output} &&

            unzip -p {input} |
            sed -e 's@<mx:collection xmlns="http://www.loc.gov/MARC21/slim">@@' |
            sed -e 's@<?xml version="1.0" encoding="UTF-8"?>@@' |
            sed -e 's@</mx:collection>@@' |
            sed -e 's@<mx:@<@g;s@</mx:@</@g' |
            tr -d '\032\033' >> {output} &&

            echo '</collection>' >> {output}""",
                          input=self.input().path,
                          preserve_whitespace=True)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml'))


class DawsonIntermediateSchema(DawsonTask):
    """
    Convert to intermediate schema via metafacture. Custom morphs and flux are
    kept in assets/124. Maps are kept in assets/maps.

    Currently broken (MF4), assets/124/124.flux: Caused by:
        org.culturegraph.mf.metamorph.MetamorphException: Error while building
        the Metamorph transformation pipeline: Error during DOM creation: Error
        parsing xml: cvc-complex-type.3.2.2: Attribute 'value' is not allowed
        to appear in element 'data'.

    """

    def requires(self):
        return DawsonFixAndCombine()

    def run(self):
        mapdir = 'file:///%s' % self.assets("maps/")
        output = shellout("""flux.sh {flux} in={input} MAP_DIR={mapdir} > {output}""",
                          flux=self.assets("124/124.flux"),
                          mapdir=mapdir,
                          input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))


class DawsonExport(DawsonTask):
    """
    Export via span-export. Hard-wired tag: DE-82.
    """
    format = luigi.Parameter(default='solr5vu3', description='export format, see span-export -list')

    def requires(self):
        return DawsonIntermediateSchema()

    def run(self):

        output = shellout("""
            cat {input} |
            span-tag -c '{{"DE-82": {{"any": {{}}}}}}' |
            span-export -o {format} > {output}""",
                          format=self.format,
                          input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='fincsolr.ndj'))
